from typing import Literal


class Options:
    logs_dir_path: str = "logs"
    file_name: str = "main.log"
    write_mode: Literal["a", "w"] = "w"
    max_MB: int = 5
    max_count: int = 10
    show_name_label: bool = False
    show_place_label: bool = False
    format_string: str = None



from logging.handlers import RotatingFileHandler
from typing import Callable
import logging
import re
from .file_handlers import get_format_string, get_rotating_file_handler
from .stream_handlers import get_stdout_handler, get_stderr_handler
from .options import Options


class DMLogger:
    options: Options = Options
    _loggers: dict = {}
    _file_handlers: dict = {}

    def __init__(
        self,
        name: str,
        level: str = "DEBUG",
        write_logs: bool = True,
        print_logs: bool = True
    ):
        if hasattr(self, '_initialized'):
            return
        self._initialized = True

        self._logger = logging.getLogger(name)
        level = logging.getLevelName(level.upper())
        self._logger.setLevel(level)
        options = DMLogger.options
        formatter = logging.Formatter(get_format_string(options), datefmt='%d-%m-%Y %H:%M:%S')

        if write_logs:
            if options.file_name in DMLogger._file_handlers:
                rotating_file_handler = DMLogger._file_handlers.get(options.file_name)
            else:
                rotating_file_handler = get_rotating_file_handler(options, formatter)
                DMLogger._file_handlers[options.file_name] = rotating_file_handler
            self._logger.addHandler(rotating_file_handler)

        if print_logs:
            self._logger.addHandler(get_stdout_handler(formatter))
            self._logger.addHandler(get_stderr_handler(formatter))

    def __new__(cls, name, *args, **kwargs):
        if name in cls._loggers:
            return cls._loggers[name]
        else:
            instance = super(DMLogger, cls).__new__(cls)
            cls._loggers[name] = instance
            return instance

    def debug(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.debug, message, **kwargs)

    def info(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.info, message, **kwargs)

    def warning(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.warning, message, **kwargs)

    def error(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.error, message, **kwargs)

    def critical(self, message: any = None, **kwargs) -> None:
        self._log(self._logger.critical, message, **kwargs)

    @staticmethod
    def _log(level_func: Callable, message: any, **kwargs) -> None:
        message = "-- " + str(message) if not (message is None) else ""
        if kwargs:
            dict_string = re.sub(r"'(\w+)':", r"\1:", str(kwargs))
            message = f"{dict_string} {message}"
        level_func(message, stacklevel=3)



from logging.handlers import RotatingFileHandler
import logging
import os
from .options import Options

default_format_string = "%(asctime)s.%(msecs)03d [%(levelname)s] [%(name)s] (%(module)s.%(funcName)s:%(lineno)d) %(message)s"


def get_rotating_file_handler(options: Options, formatter: logging.Formatter) -> RotatingFileHandler:
    logs_dir_path = options.logs_dir_path or "logs"
    logs_dir_path = os.path.normpath(logs_dir_path)
    if not os.path.exists(logs_dir_path):
        os.makedirs(logs_dir_path)
    log_path = os.path.join(logs_dir_path, options.file_name)
    is_exists_log = os.path.exists(log_path)
    file_handler = RotatingFileHandler(log_path, maxBytes=options.max_MB * 1024 * 1024, backupCount=options.max_count,
                                       encoding="utf-8")
    if options.write_mode == "w" and is_exists_log:
        file_handler.doRollover()
    file_handler.setFormatter(formatter)
    return file_handler


def get_format_string(options: Options) -> str:
    format_string = options.format_string or default_format_string
    if format_string != default_format_string:
        return format_string
    if not options.show_name_label:
        format_string = format_string.replace("[%(name)s] ", "")
    if not options.show_place_label:
        format_string = format_string.replace("(%(module)s.%(funcName)s:%(lineno)d) ", "")
    return format_string
